/*
 * Dimmer for RGB LEDs
 * Lukas Fssler, 2016-02-03
 * To be compiled using MikroC for PIC
 * See www.soldernerd.com for description, schematics, Eagle files, photos and more
 */

 /******************************************************************************
 * Standard Data Types                                                        *
 ******************************************************************************/
#define uint8_t unsigned short int
#define int8_t short int
#define uint16_t unsigned int
#define int16_t int
#define uint32_t unsigned long int
#define int32_t long int

/******************************************************************************
 * I2C                                                                        *
 ******************************************************************************/
//Slave address and mask
#define I2C_SLAVE_ADDRESS 0b11100000 /*0x70*/
#define I2C_SLAVE_ADDRESS_MASK 0b11111111 /*Do not mask any bits*/

//Struct
#define I2C_SEND_BUFFER_SIZE 11
#define I2C_RECEIVE_BUFFER_SIZE 8
typedef struct i2c_
{
  volatile uint8_t address;
  volatile uint8_t dummy;
  volatile uint8_t data_received;
  volatile uint8_t bytes_received;
  volatile uint8_t data_sent;
  volatile uint8_t bytes_sent;
  volatile uint8_t *ptr;
  uint8_t send_buffer[I2C_SEND_BUFFER_SIZE];
  uint8_t receive_buffer[I2C_RECEIVE_BUFFER_SIZE];
} i2c_struct;

/* SEND BUFFER (i.e Master Read Data)
 * Byte 0: Status
 * Byte 1: White
 * Byte 2: Off
 * Byte 3: Brightness
 * Byte 4: Color
 * Bytes 5-6: Red
 * Bytes 7-8: Green
 * Bytes 9-10: Blue
 */


/******************************************************************************
 * Constants                                                                  *
 ******************************************************************************/
#define GLOBAL_INTERRUPT_ENABLE 0b10000000
#define INTERRUPT_ON_CHANGE_ENABLE 0b00001000
#define INTERRUPT_ON_CHANGE_FLAG 0b00000001

#define ENCODER1_A 0b00010000
#define ENCODER1_B 0b00100000
#define ENCODER1_PB 0b00001000
#define ENCODER2_A 0b00000010
#define ENCODER2_B 0b00000100
#define ENCODER2_PB 0b00000001

#define NUMBER_OF_BRIGHTNESS_LEVELS 32
#define NUMBER_OF_COLORS 24

#define RGB_NEUTRAL_RED 255
#define RGB_NEUTRAL_GREEN 90
#define RGB_NEUTRAL_BLUE 50

/******************************************************************************
 * Global variables                                                           *
 ******************************************************************************/
volatile uint8_t white;
volatile uint8_t off;
volatile uint8_t brightness;
volatile uint8_t color;
volatile uint8_t update;
volatile i2c_struct i2c_data;
uint16_t red;
uint16_t green;
uint16_t blue;

uint16_t brightness_levels[NUMBER_OF_BRIGHTNESS_LEVELS] =
{
  0,
  11,
  12,
  14,
  16,
  19,
  22,
  25,
  29,
  34,
  40,
  47,
  55,
  64,
  74,
  87,
  101,
  118,
  138,
  161,
  187,
  219,
  255,
  298,
  347,
  405,
  473,
  552,
  644,
  751,
  877,
  1023
};


uint8_t colors[NUMBER_OF_COLORS][3] =
{
  {255, 0, 0},
  {223, 32, 0},
  {191, 64, 0},
  {159, 96, 0},
  {128, 128, 0},
  {96, 159, 0},
  {64, 191, 0},
  {32, 223, 0},
  {0, 255, 0},
  {0, 223, 32},
  {0, 191, 64},
  {0, 159, 96},
  {0, 128, 128},
  {0, 96, 159},
  {0, 64, 191},
  {0, 32, 223},
  {0, 0, 255},
  {32, 0, 223},
  {64, 0, 191},
  {96, 0, 159},
  {128, 0, 128},
  {159, 0, 96},
  {191, 0, 64},
  {223, 0, 32}
};

/******************************************************************************
 * Interrupt Service Routine (ISR)                                            *
 * Read and average pot value via ADC                                         *
 ******************************************************************************/
interrupt()
{
  uint8_t tmp;
  /*
   * Interrupt on Change
   */
  if(INTCON.IOCIF)
  {
    tmp = PORTB;
    //Encoder 1 Clockwise (A falling while B high)
    //Don't change brightness when off flag is set
    if(IOCBF&ENCODER1_A && tmp&ENCODER1_B && !off)
    {
      if(brightness<NUMBER_OF_BRIGHTNESS_LEVELS-1)
        ++brightness;
      update = 1;
    }
    //Encoder 1 Counter-Clockwise (B falling while A high)
    //Don't change brightness when off flag is set
    if(IOCBF&ENCODER1_B && tmp&ENCODER1_A && !off)
    {
      if(brightness>0)
        --brightness;
      update = 1;
    }
    //Encoder 1 Pushbutton
    //Don't toggle off flag when brightness is at zero
    if(IOCBF&ENCODER1_PB && brightness)
    {
      if(off)
        off = 0;
      else
        off = 1;
      update = 1;
    }
    //Encoder 2 Clockwise (A falling while B high)
    //Don't change color when white flag or off flag is set or brightness is zero
    if(IOCBF&ENCODER2_A && tmp&ENCODER2_B && !white && !off && brightness)
    {
      ++color;
      if(color==NUMBER_OF_COLORS)
        color = 0;
      update = 1;
    }
    //Encoder 2 Counter-Clockwise (B falling while A high)
    //Don't change color when white flag or off flag is set or brightness is zero
    if(IOCBF&ENCODER2_B && tmp&ENCODER2_A && !white && !off && brightness)
    {
      --color;
      if(color>NUMBER_OF_COLORS-1)
        color = NUMBER_OF_COLORS-1;
      update = 1;
    }
    //Encoder 2 Pushbutton
    //Don't toggle between color and white when off flag is set or brightness is zero
    if(IOCBF&ENCODER2_PB && !off && brightness)
    {
      if(white)
        white = 0;
      else
        white = 1;
      update = 1;
    }
    //Clear Interrupt-on-Change Flag
    IOCBF = 0;
    return;
  }
  /*
   * I2C Interrupts
   */
  if(PIR1.SSPIF)
  {
    //Master Read Mode -> Sending data
    if(SSPSTAT.R_NOT_W)
    {
      //Master did not acknowledge -> Transmission is complete
      if(SSPCON2.ACKSTAT)
      {
        i2c_data.data_sent = 1; //Indicate that data has been sent
      }
      //Master is asking for data
      else
      {
        // Address has just been received -> First byte
        if(!SSPSTAT.D_NOT_A)
        {
          i2c_data.address = SSPBUF & 0b11111110; //Save address
          i2c_data.ptr = &i2c_data.send_buffer[0]; //Reset buffer pointer
          i2c_data.bytes_sent = 0; //Reset byte counter
        }
        // Send data
        if(i2c_data.bytes_sent<I2C_SEND_BUFFER_SIZE)
        {
          SSPBUF = *i2c_data.ptr; //Fill buffer with next byte
          ++i2c_data.bytes_sent; //Incerement byte counter
          ++i2c_data.ptr; //Increment pointer
        }
        else
        {
          SSPBUF = 0x00; //No more data available, send zeros
        }
      }
    }
    //Master Transmit Mode -> Receiving data
    else
    {
      //Stop bit
      if(SSPSTAT.P)
      {
        i2c_data.data_received = 1; //Indicate that data has been received
        SSPCON3.PCIE = 0; //Disable Stop detectection interrupts
      }
      //Not stop bit
      else
      {
        // Address, not data
        if(!SSPSTAT.D_NOT_A)
        {
          i2c_data.address = SSPBUF & 0b11111110; //Save address
          i2c_data.ptr = &i2c_data.receive_buffer[0]; //Reset buffer pointer
          i2c_data.bytes_received = 0; //Reset byte counter
          SSPCON3.PCIE = 1; //Enable Stop detectection interrupts
        }
        // Data, not address
        else
        {
          if(i2c_data.bytes_received<I2C_RECEIVE_BUFFER_SIZE)
          {
            *i2c_data.ptr = SSPBUF; //Save data
            ++i2c_data.bytes_received; //Increment byte counter
            ++i2c_data.ptr; //Increment pointer
          }
        }
      }
    }
    SSPCON1.CKP = 1; //Release SCL line
    PIR1.SSPIF = 0; //Clear SSP interrupt flag
    return;
  }
}

/******************************************************************************
 * Initialize I2C module in 7-bit slave mode                                  *
 ******************************************************************************/
void i2c_slave_init()
{
  //Initialize i2c_data
  i2c_data.address = 0;
  i2c_data.dummy = 0;
  i2c_data.data_received = 0;
  i2c_data.bytes_received = 0;
  i2c_data.data_sent = 0;
  i2c_data.ptr = &i2c_data.send_buffer[0];
  for(i2c_data.bytes_sent=0; i2c_data.bytes_sent<I2C_SEND_BUFFER_SIZE; ++i2c_data.bytes_sent)
  {
    i2c_data.send_buffer[i2c_data.bytes_sent] = i2c_data.bytes_sent;
  }
  i2c_data.send_buffer[0] = 88;
  i2c_data.bytes_sent = 0;
  for(i2c_data.bytes_received=0; i2c_data.bytes_received<I2C_RECEIVE_BUFFER_SIZE; ++i2c_data.bytes_received)
  {
    i2c_data.receive_buffer[i2c_data.bytes_received] = 0;
  }
  i2c_data.bytes_received = 0;
  //Configure SCL1 and SDA1 as inputs
  TRISC.B3 = 1; //pin 14, RC3, SCL
  TRISC.B4 = 1; //pin 15, RC4, SDA
  //Set Address
  SSPADD = I2C_SLAVE_ADDRESS;
  //Set address mask
  SSPMSK = I2C_SLAVE_ADDRESS_MASK;
  /* SSPSTAT: SSP STATUS REGISTER, p286 */
  //SSPSTAT.SMP = 0; //Enable slew rate control for high speed mode (400kHz)
  SSPSTAT.SMP = 1; //Disable slew rate control for high speed mode (400kHz), RaspberryPi uses 100kHz
  //SSPSTAT.CKE = 0; //Disable SMBus specific inputs
  SSPSTAT.CKE = 1; //Enable SMBus specific inputs
  SSPSTAT.BF = 0; //Clear buffer full status bit
  /* SSPCON1: SSP CONTROL REGISTER 1, p287 */
  SSPCON1.WCOL = 0; //Clear Write Collision Detect bit
  SSPCON1.SSPOV = 0; //Clear Receive Overflow Indicator bit
  SSPCON1.SSPEN = 0; //Disable Synchronous Serial Port for now
  SSPCON1.CKP = 1; //Release clock (do not hold CLK low)
  SSPCON1.SSPM3 = 1; //7-bit I2C Slave Mode with start and stop interrupts
  SSPCON1.SSPM2 = 1; //7-bit I2C Slave Mode with start and stop interrupts
  SSPCON1.SSPM1 = 1; //7-bit I2C Slave Mode with start and stop interrupts
  SSPCON1.SSPM0 = 0; //7-bit I2C Slave Mode with start and stop interrupts
  /* SSPCON2: SSP CONTROL REGISTER 2, p288 */
  SSPCON2.GCEN = 0; //Disable General Call
  SSPCON2.SEN = 1; //Enable Clock Stretching
  /* SSPCON3: SSP CONTROL REGISTER 3 */
  SSPCON3.PCIE = 0; //Disable Stop detectection interrupts
  SSPCON3.SCIE = 0; //Disable Start detection interrupts
  SSPCON3.BOEN = 0; //Only update buffer when SSPOV is clear
  SSPCON3.SDAHT = 1; //Minimum of 300 ns hold time on SDA after the falling edge of SCL
  SSPCON3.SBCDE = 0; //Slave bus collision interrupts are disabled
  SSPCON3.AHEN = 0; //Disable address holding
  SSPCON3.DHEN = 0; //Disable data holding
  /* Configure Interrupts, p284 */
  PIR1.SSPIF = 0; //Clear SSP Interrupt Flag
  PIE1.SSPIE = 1; //Enable MSSP interrupts
  PIR2.BCLIF = 0; //Clear MSSP Bus Collision Interrupt Flag bit
  PIE2.BCLIE = 0; //Disable MSSP Bus Collision Interrupt Enable bit
  INTCON.PEIE = 1; //Enable Peripheral Interrupts
  INTCON.GIE = 1; //Enable Interrupts
  /* Turn Serial Port on */
  SSPCON1.SSPEN = 1; //Enable Synchronous Serial Port
}



/******************************************************************************
 * init()                                                                     *
 * Initialize variables, ports/pins, timer, PWM and interrupts                *
 ******************************************************************************/
static void init()
{
  //Port Setup
  //Configure Port A
  TRISA = 0b11101111; //Output for CCP5 on RA4
  PORTA = 0b00010000; //Turn inverting fet driver at RA4 off
  //Configure Port B
  TRISB = 0b11111111;
  PORTB = 0b00000000;
  //Configure Port C
  TRISC = 0b10111101; //Output for CCP3 at RC6 and CCP2 at RC1
  PORTC = 0b01000010; //Turn inverting fet driver at RC6 and RC1 off
  //Digital or analog inputs? 0=digital, 1=analog
  ANSELA = 0b00000100; //12V_SENSE at AN2
  ANSELB = 0b00000000;
  //CCP2 on RC1, CCP3 on RC6, CCP5 on RA4
  APFCON = 0b00000000;

  //Set up 32MHz internal oscillator
  //Pages 74, 83
  OSCCON = 0b11110000;
  
  //Initialize variables
  white = 1;
  off = 0;
  brightness = 1;
  color = 0;
  update = 1;

  //Configure interrupts
  IOCBP = ENCODER1_PB | ENCODER2_PB; //Interrupt-on-Change on positive edge
  IOCBN = ENCODER1_A | ENCODER1_B | ENCODER2_A | ENCODER2_B; //Interrupt-on-Change on negative edge
  INTCON &= (~INTERRUPT_ON_CHANGE_FLAG); //Clear interrupt flag prior to enable
  IOCBF = 0; // Clear interrupt-on-change flag
  INTCON |= INTERRUPT_ON_CHANGE_ENABLE; // Interrupt on change enable bit
  INTCON |= GLOBAL_INTERRUPT_ENABLE; // Global interrupt enable bit
  
  //Initialize Timer2, Timer4 & Timer6
  TMR2 = 0x00; //Start counting at 0
  PR2 = 0xFF; //Full period of 256
  TMR4 = 85; //Start counting 120 degrees out of phase
  PR4 = 0xFF; //Full period of 256
  TMR6 = 170; //Start counting 240 degrees out of phase
  PR6 = 0xFF; //Full period of 256
  T2CON = 0b00011100; //timer 2 on, prescaler=1, postscaler=4, page 209
  T4CON = 0b00011100; //timer 4 on, prescaler=1, postscaler=4, page 209
  T6CON = 0b00011100; //timer 6 on, prescaler=1, postscaler=4, page 209

  //Initialize PWM modules CCP1 and CCP2
  CCPTMRS0 = 0b00010000; //TMR2 for CCP2, TMR4 for CCP3, page 235
  CCPTMRS1 = 0b00000010; //TMR6 for CCP5, page 235
  CCP2CON = 0b00001111; //Single output, PWM mode, active-low, page 234
  CCP3CON = 0b00001111; //Single output, PWM mode, active-low, page 234
  CCP5CON = 0b00001100; //Single output, PWM mode, page 234
  
  //Initialize I2C module
  i2c_slave_init();
}

/******************************************************************************
 * PWM_red(uint16_t dutycycle)                                                *
 * Write PWM duty cycle to corresponding registers (red)                      *
 ******************************************************************************/
static void PWM_red(uint16_t dutycycle)
{
  //Truncate to 10bits
  dutycycle &= 0b1111111111;
  //Compensate for inverting driver
  dutycycle = 1023 - dutycycle;
  //Write 8 MSB
  CCPR5L = dutycycle>>2;
  //Write 2 LSB
  CCP5CON = 0b00001100 | ((dutycycle & 0b11)<<4);
}

/******************************************************************************
 * PWM_green(uint16_t dutycycle)                                              *
 * Write PWM duty cycle to corresponding registers (green)                    *
 ******************************************************************************/
static void PWM_green(uint16_t dutycycle)
{
  //Truncate to 10bits
  dutycycle &= 0b1111111111;
  //Write 8 MSB
  CCPR2L = dutycycle>>2;
  //Write 2 LSB
  CCP2CON = (CCP2CON & 0b11001111) | ((dutycycle & 0b11)<<4);
}

/******************************************************************************
 * PWM_blue(uint16_t dutycycle)                                               *
 * Write PWM duty cycle to corresponding registers (blue)                     *
 ******************************************************************************/
static void PWM_blue(uint16_t dutycycle)
{
  //Truncate to 10bits
  dutycycle &= 0b1111111111;
  //Write 8 MSB
  CCPR3L = dutycycle>>2;
  //Write 2 LSB
  CCP3CON = (CCP3CON & 0b11001111) | ((dutycycle & 0b11)<<4);
}

static void set_rgb()
{
  uint32_t tmp;
  update = 0;
  
  if(off)
  {
    red = 0;
    green = 0;
    blue = 0;
  }
  else
  {
    if(white)
    {
      tmp = ((uint32_t) brightness_levels[brightness] * RGB_NEUTRAL_RED) >> 8;
      red = (uint16_t) tmp;
      tmp = ((uint32_t) brightness_levels[brightness] * RGB_NEUTRAL_GREEN) >> 8;
      green = (uint16_t) tmp;
      tmp = ((uint32_t) brightness_levels[brightness] * RGB_NEUTRAL_BLUE) >> 8;
      blue = (uint16_t) tmp;
    }
    else
    {
      tmp = ((uint32_t) brightness_levels[brightness] * colors[color][0]) >> 8;
      red = (uint16_t) tmp;
      tmp = ((uint32_t) brightness_levels[brightness] * colors[color][1]) >> 8;
      green = (uint16_t) tmp;
      tmp = ((uint32_t) brightness_levels[brightness] * colors[color][2]) >> 8;
      blue = (uint16_t) tmp;
    }
  }
  PWM_red(red);
  PWM_green(green);
  PWM_blue(blue);
  update = 0;
}

/******************************************************************************
 * Fill the I2C send buffer                                                   *
 ******************************************************************************/
static void fill_send_buffer()
{
  i2c_data.send_buffer[0] = 0; // Status
  i2c_data.send_buffer[1] = white; // White
  i2c_data.send_buffer[2] = off; // Off
  i2c_data.send_buffer[3] = brightness; // Brightness
  i2c_data.send_buffer[4] = color; // Color
  i2c_data.send_buffer[5] = red>>7; // Red MSB
  i2c_data.send_buffer[6] = red & 0b01111111; // Red LSB
  i2c_data.send_buffer[7] = green>>7; // Green MSB
  i2c_data.send_buffer[8] = green & 0b01111111; // Green LSB
  i2c_data.send_buffer[9] = blue>>7; // Blue MSB
  i2c_data.send_buffer[10] = blue & 0b01111111; // Blue LSB
}

/******************************************************************************
 * main()                                                                     *
 ******************************************************************************/
void main() 
{
  //Initialize everything
  init();
  
  //Enter endless loop
  while(1)
  {
    if(update)
    {
      set_rgb();
      fill_send_buffer();
    }
  }
}
